/* Author: Dr. Csaba Domokos
 * Email: domokos@in.tum.de
 * Date: 22.06.2017
 * Description: This is a sample C++ code, which can be called from matlab
 *
 * Some parts of this code has been adopted from the code written by Andrew Delong
 * 
 * Compilation: You may use the following command (in Matlab) to compile the code
 *
 *              mex -O CXXFLAGS='-Wall -fPIC' -output example example.cpp
 *
 * Sample usage: (in Matlab) 
 *               A = single(reshape(1 : 9, 3, []))
 *               B = example('sqr', A)
 * 
 */


#include <mex.h>
#include <map>
#include <string>

/*
 *  Some macros for input checking and assertion handling
 */
#define MATLAB_ASSERT_ARGCOUNT(nout, nin) \
	MATLAB_ASSERT(nlhs >= nout, "Not enough output arguments, expected " #nout); \
	MATLAB_ASSERT(nlhs <= nout, "Too many output arguments, expected " #nout); \
	MATLAB_ASSERT(nrhs >= nin,  "Not enough input arguments, expected " #nin); \
	MATLAB_ASSERT(nrhs <= nin,  "Too many input arguments, expected " #nin);
#define MATLAB_ASSERT_INTYPE(arg, type) \
	MATLAB_ASSERT(mxGetClassID(prhs[arg]) == type, "Expected " #type " for input argument " #arg);

#define MATLAB_ASSERT(expr,msg) if (!(expr)) { throw MatlabError(msg); }

struct MatlabError {
	MatlabError(const char* msg): message(msg) { }
	const char* message;
};

/*
 * Auxiliary functions making an interface for user defined functions
 */
struct FuncRegistry {
	typedef void (*Func)(int, mxArray*[], int, const mxArray*[]);
	typedef std::map<std::string, Func> LookupTable;
	static LookupTable sLookup;
	struct Entry {
		Entry(const char* name, Func ptr) { sLookup[name] = ptr; }
	};
};
FuncRegistry::LookupTable FuncRegistry::sLookup;

#define EXPORT(func) \
	extern "C" void func(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]); \
	FuncRegistry::Entry regentry_##func(#func,func); \
	extern "C" void func(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])

// Mex entry function
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs == 0 || !mxIsChar(prhs[0]))
		mexErrMsgTxt("Please use a specific function");

	char funcName[512];
	mxGetString(prhs[0], funcName, mxGetN(prhs[0])*sizeof(mxChar)+1);

	FuncRegistry::LookupTable::const_iterator it = FuncRegistry::sLookup.find(funcName);

	if (it == FuncRegistry::sLookup.end())
		mexErrMsgTxt("Specified function does not exist");

	try {
		it->second(nlhs, plhs, nrhs-1, prhs+1);
    } catch (MatlabError err) {
		mexErrMsgTxt(err.message);
    }
}
///////////////////////////////////////////////////////////////////////////
    
/*
 * You may write your own functions after this point
 * 
 * Please note that the matrices in Matlab are stored in a **column major order**
 */

/* 
 * Calculating element-wise square for an input matrix
 */
EXPORT(sqr)
{
    // check the number of input/output arguments
    MATLAB_ASSERT_ARGCOUNT(1, 1);
    
    // check the input
    MATLAB_ASSERT_INTYPE(0, mxSINGLE_CLASS);
    
    const int m = mxGetM(prhs[0]);
    const int n = mxGetN(prhs[0]);
    MATLAB_ASSERT(mxGetNumberOfDimensions(prhs[0]) == 2 and m > 1 and n > 1, "The input must be a matrix");
    
    // read the input
    float *p = (float*)mxGetData(prhs[0]);
    
    // allocate memory for the result
    plhs[0] = mxCreateNumericMatrix(m, n, mxDOUBLE_CLASS, mxREAL);
    double *p2 = (double*)mxGetData(plhs[0]);
    
    // compute the result
    for(int i = 0; i < m * n; ++i)
        p2[i] = p[i] * p[i];
}